
/*
 *    miracleshaper - Traffic shaper and prioritizer for DSLs.
 *    Copyright (C) 2007  Florian Zumbiehl <florz@florz.de>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

// The number of configuration profiles we are going to use. This is
// used by main.c:
#define PROFILE_COUNT 2

#define MODEM_MAC 0x000b238f0a03 // see below what it's used for

/* This function is called whenever a profile is being activated or
 * deactivated and is expected to do the actual reconfiguration based
 * on which profiles are currently active. Usually, it should reconfigure
 * things for the most restrictive profile that is currently active.
 *
 * Profiles are activated by a call to profile_activate() from one
 * of the classifier functions below, specifying the number of the
 * profile to activate and the number of seconds until it is supposed
 * to be deactivated (unless that timeout is updated to a different value
 * before it expires by another call to profile_activate()).
 */
static void profile_commit(void){
	int data_rate,max_queue_size;

	if(profile_flags&(1<<1)){ // for VoIP
		data_rate=DOWNSTREAM_BANDWIDTH/2;
		max_queue_size=5;
	}else if(profile_flags&(1<<0)){ // for SSH
		data_rate=DOWNSTREAM_BANDWIDTH*9/10;
		max_queue_size=50;
	}else{
		data_rate=DOWNSTREAM_BANDWIDTH;
		max_queue_size=0;
	}
	prioqueue_set_data_rate(downstreamq,data_rate);
	// The unit of the max_queue_size argument is the MTU size.
	// So, this does specify the "number of full-sized packets",
	// if you will. But with smaller packets, the amount of data
	// is what counts, not the number of packets. At least as
	// long as you don't hit DOWNSTREAM_QUEUE_MAX_PACKETS.
	prioqueue_set_max_queue_size(downstreamq,max_queue_size);
}

/* The following two functions are called for each packet that is about to be
 * enqueued in the upstream or downstream queue, respectively, and is expected
 * to return either a priority (a value of 0 or greater) or one of the values
 * PRIO_DROP, which causes the packet to be silently dropped, or
 * PRIO_IMMEDIATE, which causes it to be forwarded immediately, without any
 * queueing delay. In the latter case, the packet still is accounted for as
 * having used bandwidth and thus causes any packets that are put into the
 * queue to be delayed accordingly.
 *
 * For dissecting the packets, you can use the functions from pkt.c and pkt.h,
 * see there for documentation.
 */

static int upstream_classify(struct pkt *pkt){
	long eth_proto;

	pkt_pinit(pkt);
	if((eth_proto=pkt_eth_get_proto(pkt))<0)return PRIO_DROP;
	if(eth_proto==ETH_P_PPP_DISC){
		return 0; // PPPoE discovery
	}
	if(eth_proto==ETH_P_PPP_SES){
		long ppp_proto;

		if(!pkt_eth_nextlayer(pkt))return PRIO_DROP;
		if(!pkt_pppoeses_nextlayer(pkt))return PRIO_DROP;
		if((ppp_proto=pkt_ppp_get_proto(pkt))<0)return PRIO_DROP;
		if(!pkt_ppp_nextlayer(pkt))return PRIO_DROP;
		if(ppp_proto==PPP_P_IP){
			int ip_proto;
			int ip_tos;

			if((ip_proto=pkt_ip_get_proto(pkt))<0)return PRIO_DROP;
			if((ip_tos=pkt_ip_get_tos(pkt))<0)return PRIO_DROP;
			if(!pkt_ip_nextlayer(pkt))return PRIO_DROP;
			if(ip_proto==IPPROTO_ICMP){
				//return 1; // This can be used for easy testing with IPv4 ICMP pings.
			}
			if(ip_proto==IPPROTO_TCP){
				if(ip_tos==0x10){
					// IP ToS of 0x10 is what OpenSSH uses for interactive sessions by default.
					profile_activate(0,20); // activate SSH profile for the next 20 seconds.
					return 2;
				}
				if(!pkt_tcp_nextlayer(pkt))return PRIO_DROP;
				if(pkt_get_plen(pkt)==0){
					return 2; // IPv4 TCP packet without payload, likely an ACK or SYN.
				}
			}
			if(ip_proto==IPPROTO_UDP&&ip_tos==0xa0){
				long udp_dstport;

				if((udp_dstport=pkt_udp_get_dstport(pkt))<0)return PRIO_DROP;
				if(udp_dstport!=5060){
					// This is for IPv4 UDP packets with an IP ToS field of 0xa0 (which
					// happens to be what my VoIP provider uses) not going to
					// port 5060 - the latter is because asterisk marks not only
					// the RTP packets, but also the SIP packets, in particular the
					// regular REGISTER packets, which I don't want to halve my
					// bandwidth every once a minute.
					profile_activate(1,3); // activate VoIP profile for the next 3 seconds
					return 1;
				}
			}
			return 3; // any other IPv4 traffic within a PPPoE/PPP connection
		}
		if(ppp_proto==PPP_P_IPV6){
			int ip6_proto;
			int ip6_tc;

			if((ip6_proto=pkt_ip6_get_proto(pkt))<0)return PRIO_DROP;
			if((ip6_tc=pkt_ip6_get_tc(pkt))<0)return PRIO_DROP;
			if(!pkt_ip6_nextlayer(pkt))return PRIO_DROP;
			if(ip6_proto==IPPROTO_TCP){
				if(ip6_tc==0x10){
					// IP ToS of 0x10 is what SSH uses for interactive sessions by default.
					profile_activate(0,20); // activate SSH profile for the next 20 seconds.
					return 2;
				}
				if(!pkt_tcp_nextlayer(pkt))return PRIO_DROP;
				if(pkt_get_plen(pkt)==0){
					// TCP without payload, this time within of IPv6
					return 2;
				}
			}
			return 3; // any other IPv6 traffic within a PPPoE/PPP connection
		}
		// Any other PPP frames - unless you are also using IPX or some other strange
		// protocols over this, it must be LCP, IPCP or IP6CP frames, which should be
		// forwarded ASAP in order to avoid LCP echo timeouts, for example.
		return 0;
	}
	{
		long long eth_dst;

		if((eth_dst=pkt_eth_get_dst(pkt))<0)return PRIO_DROP;
		if(eth_dst==MODEM_MAC){
			// This is in case you also want to access your modem via IP.
			// Maybe you will also have to let ARP through somehow - I don't
			// since my modem doesn't answer ARP requests anyway.
			return 4;
		}
	}
	return PRIO_DROP;
}

static int downstream_classify(struct pkt *pkt){
	long eth_proto;

	pkt_pinit(pkt);
	if((eth_proto=pkt_eth_get_proto(pkt))<0)return PRIO_DROP;
	if(eth_proto==ETH_P_PPP_DISC){
		return PRIO_IMMEDIATE; // PPPoE discovery
	}
	if(eth_proto==ETH_P_PPP_SES){
		long ppp_proto;

		if(!pkt_eth_nextlayer(pkt))return PRIO_DROP;
		if(!pkt_pppoeses_nextlayer(pkt))return PRIO_DROP;
		if((ppp_proto=pkt_ppp_get_proto(pkt))<0)return PRIO_DROP;
		if(!pkt_ppp_nextlayer(pkt))return PRIO_DROP;
		if(ppp_proto==PPP_P_IP){
			int ip_proto;
			int ip_tos;

			if((ip_proto=pkt_ip_get_proto(pkt))<0)return PRIO_DROP;
			if((ip_tos=pkt_ip_get_tos(pkt))<0)return PRIO_DROP;
			if(!pkt_ip_nextlayer(pkt))return PRIO_DROP;
			if(ip_proto==IPPROTO_ICMP){
				//return PRIO_IMMEDIATE; // This can be used for easy testing with IPv4 ICMP pings.
			}
			if(ip_proto==IPPROTO_TCP){
				if(ip_tos==0x10){
					// IP ToS of 0x10 is what OpenSSH uses for interactive sessions by default.
					return 0;
				}
				if(!pkt_tcp_nextlayer(pkt))return PRIO_DROP;
				if(pkt_get_plen(pkt)==0){
					return 0; // IPv4 TCP packet without payload, likely an ACK or SYN.
				}
			}
			if(ip_proto==IPPROTO_UDP&&ip_tos==0xa0){
				return PRIO_IMMEDIATE; // IP ToS 0xa0 is used by my VoIP provider.
			}
			return 1; // any other IPv4 traffic within a PPPoE/PPP connection
		}
		if(ppp_proto==PPP_P_IPV6){
			int ip6_proto;

			if((ip6_proto=pkt_ip6_get_proto(pkt))<0)return PRIO_DROP;
			if(!pkt_ip6_nextlayer(pkt))return PRIO_DROP;
			if(ip6_proto==IPPROTO_TCP){
				if(!pkt_tcp_nextlayer(pkt))return PRIO_DROP;
				if(pkt_get_plen(pkt)==0){
					// TCP without payload, this time within of IPv6
					return 0;
				}
			}
			return 1; // any other IPv6 traffic within a PPPoE/PPP connection
		}
		// Any other PPP frames - unless you are also using IPX or some other strange
		// protocols over this, it must be LCP, IPCP or IP6CP frames, which should be
		// forwarded ASAP in order to avoid LCP echo timeouts, for example.
		return PRIO_IMMEDIATE;
	}
	{
		long long eth_src;

		if((eth_src=pkt_eth_get_src(pkt))<0)return PRIO_DROP;
		if(eth_src==MODEM_MAC){
			// This is in case you also want to access your modem via IP.
			return 2;
		}
	}
	return PRIO_DROP;
}

